package ap.basetypes;

import ap.util.Debug

object LeftistHeap {

  private[basetypes] val AC = Debug.AC_BASE_TYPE

  def EMPTY_HEAP[T <% Ordered[T]] : LeftistHeap[T] = new EmptyHeap[T]

  private[basetypes] def node[T <% Ordered[T]](element : T) : Node[T] = {
    val empty = LeftistHeap.EMPTY_HEAP[T]
    new Node(element, empty, empty, empty)
  }

  private[basetypes] def node[T <% Ordered[T]](element : T,
                                               a : LeftistHeap[T],
                                               b : LeftistHeap[T],
                                               empty : LeftistHeap[T]) : Node[T] =
    if (a.rightHeight <= b.rightHeight)
      new Node (element, b, a, empty)
    else
      new Node (element, a, b, empty)
}

/**
 * This class implements the leftist heap, see &quot;Functional Data
 * Structures&quot; by Chris Okasaki
 */
abstract class LeftistHeap[T <% Ordered[T]] extends Collection[T] {

   /**
    * @return true iff this heap is empty
    */
   def isEmpty : Boolean

   /**
    * Length of the right spine, i.e. the length of the path from the
    * root to rightmost leaf
    */
   def rightHeight : Int

   /**
    * @return the minimum element of this heap, or raise an exception iff this
    * heap is empty (<code>isEmpty()==true</code>)
    */
   def findMin : T

   /**
    * Remove the minimum element from this heap, or raise an exception iff this
    * heap is empty (<code>isEmpty()==true</code>)
    * @return a heap that contains all elements of this heap except
    * for the minimum element
    */
   def deleteMin : LeftistHeap[T]
   
   /**
    * Add an element to this heap object
    * @param element The element to be added
    * @return a heap that contains all elements of this heap, and
    * additionally <code>element</code>
    */
   def insert(element : T) : LeftistHeap[T]

   /**
    * Add multiple elements to this heap object. We keep this method protected,
    * because otherwise one could use it to insert heaps that are sorted
    * differently
    * @param h a heap containing the elements to be added
    * @return a heap that contains all elements of this heap, and
    * additionally all objects from <code>h</code>
    */
   protected[basetypes] def insertHeap(h : LeftistHeap[T]) : LeftistHeap[T]

   /**
    * Add multiple elements to this heap object
    * @param elements the elements to be added
    * @return a heap that contains all elements of this heap, and
    * additionally all objects from <code>elements</code>
    */
   def insertIt(elements : Iterator[T]) : LeftistHeap[T] = {
        // Use bottom-up strategy to compose new heap in O(n)

        val s = new scala.collection.mutable.Stack[LeftistHeap[T]]
        s.push(this)
        while (elements.hasNext) {
            var h : LeftistHeap[T] = LeftistHeap.node(elements.next)
            while (!s.isEmpty && h.size >= s.top.size) h = h.insertHeap(s.pop)
            s.push(h)
        }
        var res : LeftistHeap[T] = s.pop
        while (!s.isEmpty) res = res.insertHeap(s.pop);
        res
   }

   /**
    * @return an iterator that returns all elements of this heap
    */
   def unsortedIterator : Iterator[T] = new UnsortedIterator (this)
     
   /**
    * @return an iterator that returns all elements of this heap in
    * increasing order
    */
   def sortedIterator : Iterator[T] = new SortedIterator (this)

   /**
    * Remove all elements of this heap which are <code>equal</code>
    * to <code>element</code>.
    * @return heap that has all occurrences of <code>element</code>
    * removed
    */
   def removeAll(element : T) : LeftistHeap[T]

   /////////////////////////////////////////////////////////////////////////////
   
   def elements : Iterator[T] = unsortedIterator
   
   def +(el : T) : LeftistHeap[T] = this.insert(el)
   
   def ++(els : Iterator[T]) : LeftistHeap[T] = this.insertIt(els)
   
   def ++(els : Iterable[T]) : LeftistHeap[T] = this.insertIt(els.elements)
   
   /////////////////////////////////////////////////////////////////////////////
   
   override def toString : String = {
      val it=this.unsortedIterator
      val str = new StringBuffer("[")
      while (it.hasNext) {
        str.append(""+it.next)
        if (it.hasNext) {
          str.append(",")
        }
      }        
      str.append("]")
      str.toString()
   }

}

 

class UnsortedIterator[A] extends Iterator[A] {
  
  private val remainder = new scala.collection.mutable.Stack[LeftistHeap[A]]

  def this(heap : LeftistHeap[A]) = {
    this()
    push(heap)
  }

  private def push(heap : LeftistHeap[A]) : Unit =
    if (!heap.isEmpty) remainder.push(heap)

  def hasNext : Boolean = !remainder.isEmpty

  def next : A = {
    remainder.pop match {
      case Node(data, left, right, _) => {
        // descend in right-first order, this helps to keep the stack small
        push ( left )
        push ( right )
        data
      }
    }
  }
}



/**
 * Class for iterating the elements of a heap in increasing order
 */
class SortedIterator[A](var remainder : LeftistHeap[A]) extends Iterator[A] {

  def hasNext : Boolean = !remainder.isEmpty

  def next : A = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre ( LeftistHeap.AC, !remainder.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
     
    val res = remainder.findMin
    remainder = remainder.deleteMin
    res
  }
}



  /**
   * Use this class to construct new heaps
   */
  case class EmptyHeap[T <% Ordered[T]] extends LeftistHeap[T] {

    /**
     * Length of the right spine, i.e. the length of the path from the
     * root to rightmost leaf
     */
    val rightHeight : Int = 0

    /**
     * @return the number of elements this heap holds
     */
    val size : Int = 0

    /**
     * @return true iff this heap is empty
     */
    override def isEmpty : Boolean = true

    /**
     * Add an element to this heap object
     * @param element The element to be added
     * @return a heap that contains all elements of this heap, and
     * additionally <code>element</code>
     */
    def insert(element : T) : LeftistHeap[T] =
      LeftistHeap.node (element, this, this, this)

    /**
     * Add multiple elements to this heap object
     * @param h a heap containing the elements to be added
     * @return a heap that contains all elements of this heap, and
     * additionally all objects from <code>h</code>
     */
    protected[basetypes] def insertHeap(h : LeftistHeap[T]) : LeftistHeap[T] = h

    /**
     * @return the minimum element of this heap, or null iff this heap
     * is empty (<code>isEmpty()==true</code>)
     */
    def findMin : Nothing = throw new NoSuchElementException

    /**
     * Remove the minimum element from this heap
     * @return a heap that contains all elements of this heap except
     * for the minimum element
     */
    def deleteMin : LeftistHeap[T] = throw new NoSuchElementException

    /**
     * Remove all elements of this heap which are <code>equal</code>
     * to <code>element</code>.
     * @return heap that has all occurrences of <code>element</code>
     * removed
     */
    def removeAll(element : T) : LeftistHeap[T] = this

  }



 
/**
 * Class for non-empty heaps. We also keep a reference to the empty heap to
 * avoid creating new objects (ugly ... there should really be explicit 
 * Node-classes for nodes with no or only one child)
 */
case class Node[T <% Ordered[T]](data : T,
                                 left : LeftistHeap[T],
                                 right : LeftistHeap[T],
                                 empty : LeftistHeap[T]) extends LeftistHeap[T] {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(LeftistHeap.AC,
                   // the heap property
                   (left.isEmpty || data <= left.findMin) &&
                   (right.isEmpty || data <= right.findMin) &&
                   // the property of a leftist heap
                   right.rightHeight <= left.rightHeight)
  //////////////////////////////////////////////////////////////////////////////
   
  /**
   * Length of the right spine, i.e. the length of the path from the
   * root to rightmost leaf
   */
  val rightHeight : Int = right.rightHeight + 1
  
  val size : Int = left.size + right.size + 1

  /**
   * @return true iff this heap is empty
   */
  override def isEmpty : Boolean = false
   
  /**
   * Add an element to this heap object
   * @param element the element to be added
   * @return a heap that contains all elements of this heap, and
   * additionally <code>element</code>
   */
  def insert(element : T) : LeftistHeap[T] =
    if (element <= data)
      LeftistHeap.node(element, this, empty, empty)
    else
      LeftistHeap.node(data, left, right.insert(element), empty)

  /**
   * Add multiple elements to this heap object
   * @param h a heap containing the elements to be added
   * @return a heap that contains all elements of this heap, and
   * additionally all objects from <code>h</code>
   */
  protected[basetypes] def insertHeap(h : LeftistHeap[T]) : LeftistHeap[T] = h match {
    case _ : EmptyHeap[T] => this
    case Node(hdata, hleft, hright, _) =>
      if (data <= hdata)
        LeftistHeap.node(data, left, right.insertHeap(h), empty)
      else
        LeftistHeap.node(hdata, hleft, this.insertHeap(hright), empty)
  }

  /**
   * @return the minimum element of this heap, or null iff this heap
   * is empty (<code>isEmpty()==true</code>)
   */
  def findMin : T = data

  /**
   * Remove the minimum element from this heap
   * @return a heap that contains all elements of this heap except
   * for the minimum element
   */
  def deleteMin : LeftistHeap[T] = left.insertHeap(right)

  /**
   * Remove all elements of this heap which are <code>equal</code>
   * to <code>element</code>.
   * @return heap that has all occurrences of <code>element</code>
   * removed
   */
  def removeAll(element : T) : LeftistHeap[T] = {
    val c = data.compareTo ( element )

    if ( c > 0 ) {
      this
    } else {
      val newLeft = left.removeAll(element)
      val newRight = right.removeAll(element)
            
      if ( c == 0 && data == element ) {
        newLeft.insertHeap ( newRight )
      } else {
        LeftistHeap.node(data, newLeft, newRight, empty)
      }
    }
  }
}

