package ap.proof;

import ap.terfor.{Formula, TermOrder, Conjunction, Quantifier, NegatedConjunctions}
import ap.proof.goal.Goal
import ap.util.{Debug, Seqs}

object ConstraintSimplifier {
  
  val AC = Debug.AC_CONSTRAINT_SIMPLIFIER  
  
  val NO_SIMPLIFIER : ConstraintSimplifier = new ConstraintSimplifier {
    def apply(f : Formula, order : TermOrder) : Formula = f
  }
  
  val ALL_SIMPLIFIER : ConstraintSimplifier =
    new SimpleSimplifier(true)

  val QUANTIFICATION_SIMPLIFIER : ConstraintSimplifier =
    new SimpleSimplifier(false)

}

abstract class ConstraintSimplifier {
  /**
   * Simplify the formula <code>f</code>
   */
  def apply(f : Formula, order : TermOrder) : Formula
}

class SimpleSimplifier(groundSimplifications : Boolean)
      extends ConstraintSimplifier {
  
  private lazy val prover =
    new ExhaustiveProver(ConstraintSimplifier.NO_SIMPLIFIER)

  private val output = false
    
  private def print(str : => String) : unit = (if (output) Predef.print(str))
  private def println(str : => String) : unit = (if (output) Predef.println(str))

  def apply(f : Formula, order : TermOrder) : Formula = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ConstraintSimplifier.AC, order isSortingOf f)
    ////////////////////////////////////////////////////////////////////////////

    // TODO: check that the formula actually can be handled

    firstQuantifier(f) match {
    case Some(Quantifier.ALL) => simplify(f, order)
    case Some(Quantifier.EX) => negSimplify(f, order)
    case None if (groundSimplifications) => f match {
      case f : Conjunction if (f.size > 1) => negSimplify(f, order)
      case _ => f
    }
    case _ => f
    }
  }

  private def simplify(f : Formula, order : TermOrder) : Formula = {
    print("Simplify: " + f + " (positive) ")
    val res = prover(Goal(f, order)).closingConstraint
    println("done")
    res
  }

  private def negSimplify(f : Formula, order : TermOrder) : Formula = {
    print("Simplify: " + f + " (negative) ")
    val goal = Goal(Conjunction.negate(f, order), order)
    val res = Conjunction.negate(prover(goal).closingConstraint, order)    
    println("done")
    res
  }
  
  /**
   * Find an outermost quantifier in <code>f</code> and return it, or return
   * <code>None</code> if no quantifier was found
   */
  private def firstQuantifier(f : Formula) : Option[Quantifier] = f match {
    case f : Conjunction =>
      if (f.isDivisibility || f.isNonDivisibility)
        None
      else if (f.quans.isEmpty)
        firstQuantifier(f.negatedConjs)
      else
        Some(f.quans.last)
    case f : NegatedConjunctions =>
      for (q <- Seqs.some(for (conj <- f.elements) yield firstQuantifier(conj)))
      yield q.dual
    case _ => None
  }
  
}
