package ap.proof;

import ap.proof.goal.Goal
import ap.util.Logic

/**
 * A prover that tries to construct an exhaustive proof for a given goal. The
 * prover tries to optimise by early stopping the expansion of the proof tree
 * if it is detected that a certain subtree can never yield a satisfiable
 * closing constraint
 */
class ExhaustiveProver(simplifier : ConstraintSimplifier) {

  private val ptf = new SimpleProofTreeFactory(true, simplifier)
   
  def apply(goal : Goal) : ProofTree = {
    var tree : ProofTree = goal
    var cont : Boolean = true
    while (cont) {
      val (newTree, newCont) = expandProof(tree, 1)
      tree = newTree
      cont = newCont
    }
    tree
  }
    
  /**
   * Try to expand each goal of <code>tree</code> by applying <code>limit</code>
   * steps. The result is a pair consisting of the new proof tree and a boolean
   * that tells whether it was possible to apply any steps
   */
  private def expandProof(tree : ProofTree, limit : Int) : (ProofTree, Boolean) =
    if (limit > 0 && tree.stepMeaningful) tree match {
      
      case goal : Goal => (expandProof(goal.step(ptf), limit - 1) _1, true)
      
      case tree : ProofTreeOneChild => {
        val (newSubtree, stepWasPossible) = expandProof(tree.subtree, limit)
        (tree update newSubtree, stepWasPossible)
      }
      
      case tree : AndTree => {
        var stepWasPossible : Boolean = false
        val newSubtrees = for (subtree <- tree.subtrees) yield {
          val (newSubtree, swp) = expandProof(subtree, limit)
          stepWasPossible = stepWasPossible || swp
          newSubtree
        }
        (tree update newSubtrees, stepWasPossible)
      }
      
    } else {
      (tree, false)
    }
}
