package ap.proof;

import ap.terfor.{Quantifier, TermOrder, ConstantTerm, VariableTerm, Formula,
                  Conjunction}
import ap.terfor.substitutions.{Substitution, ConstantSubst, IdentitySubst,
                                ComposeSubsts}
import ap.util.{PlainRange, Debug, Seqs}

object QuantifiedTree {
  
  private val AC = Debug.AC_PROOF_TREE
  
  def apply(subtree : ProofTree,
            quan : Quantifier,
            quantifiedConstants : Set[ConstantTerm],
            order : TermOrder,
            subtreeOrder : TermOrder,
            simplifier : ConstraintSimplifier) : ProofTree = subtree match {
    case subtree : QuantifiedTree if (subtree.quan == quan) =>
      QuantifiedTree(subtree.subtree,
                     quan,
                     quantifiedConstants ++ subtree.quantifiedConstants,
                     order, subtree.subtreeOrder,
                     simplifier)
    case subtree : AndTree if (quan == Quantifier.ALL) =>
      AndTree(for (t <- subtree.subtrees)
              yield QuantifiedTree(t, quan, quantifiedConstants, order,
                                   subtreeOrder, simplifier),
              order, simplifier)
    case subtree : WeakenTree if (quan == Quantifier.EX) => {
      println("New case!")
      WeakenTree(QuantifiedTree(subtree.subtree,
                                quan, quantifiedConstants,
                                order, subtreeOrder, simplifier),
                 quantify(subtree.disjunct, quan, quantifiedConstants,
                          order, subtreeOrder, simplifier),
                 order)
    }
    case subtree : WeakenTree if (Seqs.disjoint(quantifiedConstants,
                                                subtree.disjunct.constants)) =>
      WeakenTree(QuantifiedTree(subtree.subtree,
                                quan, quantifiedConstants,
                                order, subtreeOrder, simplifier),
                 order sort subtree.disjunct,
                 order)
    case _ => 
      new QuantifiedTree(subtree, quan, quantifiedConstants, order, subtreeOrder,
                         simplifier)
  }
 
  private def quantify(f : Formula,
                       quan : Quantifier, quantifiedConstants : Set[ConstantTerm],
                       outputOrder : TermOrder, inputOrder : TermOrder,
                       simplifier : ConstraintSimplifier) : Formula = {
    var constantSubst : Substitution = new IdentitySubst(inputOrder)
    for ((c, i) <- quantifiedConstants.elements.zipWithIndex) {
      val nextSubst = new ConstantSubst(c, VariableTerm(i), inputOrder)
      constantSubst = ComposeSubsts(constantSubst, nextSubst, inputOrder)
    }
    
    val quans : Seq[Quantifier] =
      (for (_ <- PlainRange(quantifiedConstants.size)) yield quan)    
    
    val constraintWithVars = constantSubst(f)
    val nonSimplified =
      (Conjunction.quantify(quans, constraintWithVars, inputOrder) sortBy outputOrder)
      
    simplifier(nonSimplified, outputOrder)    
  }
  
}

/**
 * <code>ProofTreeOneChild</code> that quantifies a set of constants in
 * the closing constraint of its <code>subtree</code>
 */
class QuantifiedTree private (val subtree : ProofTree,
                              val quan : Quantifier,
                              val quantifiedConstants : Set[ConstantTerm],
                              protected val order : TermOrder,
                              protected val subtreeOrder : TermOrder,
                              simplifier : ConstraintSimplifier)
      extends ProofTreeOneChild {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(QuantifiedTree.AC,
                   !quantifiedConstants.isEmpty &&
                   (subtree match {
                     case subtree : QuantifiedTree => subtree.quan != this.quan
                     case subtree : AndTree => this.quan != Quantifier.ALL
                     case subtree : WeakenTree if (this.quan == Quantifier.EX) =>
                       false
                     case subtree : WeakenTree if (this.quan == Quantifier.ALL) =>
                       !(quantifiedConstants intersect subtree.disjunct.constants).isEmpty
                     case _ => true
                    }))
  //////////////////////////////////////////////////////////////////////////////

  lazy val closingConstraint : Formula =
    QuantifiedTree.quantify(subtree.closingConstraint, quan, quantifiedConstants,
                            order, subtreeOrder, simplifier)
  
  lazy val stepMeaningful : Boolean = subtree.stepMeaningful

  def update(newSubtree : ProofTree) : ProofTree =
    QuantifiedTree(newSubtree, quan, quantifiedConstants, order, subtreeOrder,
                   simplifier)

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = {
    "QuantifiedTree (" + quan + ", " + quantifiedConstants + ") (" +
      closingConstraint + ") {\n" + subtree + "\n}"
  }

}
