package ap.proof.goal;

import ap.terfor.{Formula, Conjunction, Quantifier, ConstantTerm}
import ap.terfor.arithconj.ArithConj
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.substitutions.VariableSubst
import ap.util.{Debug, PlainRange}

object BetaFormulaTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  def isCoveredFormula(f : Conjunction) : Boolean =
    f.size > 1 && f.quans.isEmpty

}


class BetaFormulaTask(_formula : Conjunction, _age : Int)
      extends FormulaTask(_formula, _age) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(BetaFormulaTask.AC,
                   !formula.isTrue && !formula.isFalse &&
                   !formula.isLiteral && !formula.isNegatedConjunction)
  //////////////////////////////////////////////////////////////////////////////
      
  val priority : Int = age

  /**
   * Perform the actual task (whatever needs to be done with <code>formula</code>)
   */
  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree =
    if (formula.arithConj.isTrue) {
      // split to handle the complex conjuncts
      applyBeta(goal, ptf)
    } else {
      // cut over one of the literals
      applyCut(goal, ptf)
    }

  private def applyBeta(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, formula.negatedConjs.size > 1)
    ////////////////////////////////////////////////////////////////////////////
      
    val subtrees = for (negConj <- formula.negatedConjs)
                   yield ptf.updateGoal(FormulaTask(negConj.negate, goal.age), goal)
    ptf.and(subtrees, goal.order)
  }

  private def applyCut(goal : Goal, ptf : ProofTreeFactory)
                                                                : ProofTree = {
    val order = goal.order    
    val (cutLiteral, rem) = selectCutLiteral(formula)
    
    val firstTree = {
      // assume that the cut-literal is true
      val updatedFacts = Conjunction.conj(Array(goal.facts, cutLiteral), order)
      val newTask = FormulaTask(rem, goal.age)
      ptf.updateGoal(//updatedFacts,    // normal tableau-splitting for the time being
                     newTask ++ List(FactsNormalisationTask),
                     goal)
    }
    
    val secondTree = {
      // assume that the cut-literal is false
      val updatedFacts = Conjunction.conj(Array(goal.facts, cutLiteral.negate), order)
      ptf.updateGoal(updatedFacts, List(FactsNormalisationTask), goal)
    }
    
    ptf.and(Array(secondTree, firstTree), order)    
  }
  
  /**
   * Given a conjunction of formulas, select one of the formulas
   * as literal to cut over. The result is the pair of the selected formula and
   * the remaining <code>Conjunction</code>
   */
  private def selectCutLiteral(conj : Conjunction) : (Conjunction, Conjunction) = {
    ////////////////////////////////////////////////////////////////////////////
    // TODO
    Debug.assertPre(BetaFormulaTask.AC, !conj.arithConj.isTrue)
    ////////////////////////////////////////////////////////////////////////////

    val (sel, rem) = selectCutLiteral(conj.arithConj)
    (Conjunction.conj(sel, conj.order), conj.updateArithConj(rem, conj.order))
  }

  private def selectCutLiteral(ac : ArithConj) : (ArithConj, ArithConj) =
    if (ac.positiveEqs.isEmpty) {
      val (sel, rem) = selectCutLiteral(ac.negativeEqs)
      (ArithConj.conj(sel, ac.order), ac.updateNegativeEqs(rem, ac.order))
    } else {
      val (sel, rem) = selectCutLiteral(ac.positiveEqs)
      (ArithConj.conj(sel, ac.order), ac.updatePositiveEqs(rem, ac.order))
    }

  private def selectCutLiteral(eqs : EquationConj) : (EquationConj, EquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (EquationConj(eqs(eqs.size - 1), eqs.order),
     EquationConj(eqs.slice(0, eqs.size - 1), eqs.order))
  }

  private def selectCutLiteral(eqs : NegEquationConj)
                                      : (NegEquationConj, NegEquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (NegEquationConj(eqs(eqs.size - 1), eqs.order),
     NegEquationConj(eqs.slice(0, eqs.size - 1), eqs.order))
  }
  
  /**
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction) : FormulaTask =
    new BetaFormulaTask(f, age)

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  protected def isCoveredFormula(f : Conjunction) : Boolean =
    BetaFormulaTask.isCoveredFormula(f)

}
