package ap.proof.goal;

import ap.basetypes.IdealInt
import ap.terfor.{Formula, Conjunction, Quantifier, ConstantTerm, VariableTerm}
import ap.terfor.substitutions.VariableSubst
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.linearcombination.LinearCombination
import ap.util.{Debug, PlainRange}

import scala.collection.mutable.ArrayBuffer

object ExQuantifierTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  def isCoveredFormula(f : Conjunction) : Boolean =
    f.quans.lastOption == Some(Quantifier.EX)

}

class ExQuantifierTask(_formula : Conjunction, _age : Int)
      extends QuantifierTask(_formula, _age) {

  val priority : Int = 1 + age

  protected val constantNameBase : String = "ex_"
    
  protected def newEliminatedConstants(goal : Goal,
                                       newConstants : Iterable[ConstantTerm])
                                               : Set[ConstantTerm] = Set.empty

  /**
   * Perform the actual task (whatever needs to be done with <code>formula</code>)
   */
  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree =
    if (formula.isDivisibility)
      caseDistinction(goal, ptf)
    else
      instantiateWithConstants(goal, ptf)    
  
  /**
   * Transform a formula <code>EX (n*_0 + t = 0)</code> in the succedent into
   * a conjunction
   * <code>ALL (n*_0 + t + 1 != 0) /\ ALL (n*_0 + t + 2 != 0) /\ ...</code>
   * in the succedent
   */
  private def caseDistinction(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val lc = formula.arithConj.positiveEqs(0)
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertInt(ExQuantifierTask.AC,
                    !lc.isEmpty && lc.leadingTerm == VariableTerm(0))
    ////////////////////////////////////////////////////////////////////////////
    
    val coeff : IdealInt = lc.leadingCoeff
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertInt(ExQuantifierTask.AC, coeff.signum > 0)
    ////////////////////////////////////////////////////////////////////////////
    
    val order = goal.order
    
    val cases = new ArrayBuffer[Formula]
    
    var offset : IdealInt = IdealInt.ONE
    while (offset < coeff) {
      val lcWithOffset = LinearCombination.sum(Array((IdealInt.ONE, lc),
                                                     (offset, LinearCombination.ONE)),
                                               order)
      cases += NegEquationConj(lcWithOffset, order)
      offset = offset + IdealInt.ONE
    }
    
    val caseDistinction = Conjunction.quantify(Array(Quantifier.ALL),
                                               Conjunction.conj(cases, order),
                                               order)
    
    ptf.updateGoal(FormulaTask(caseDistinction, goal.age), goal)
  }
   
  /**
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction) : FormulaTask =
    new ExQuantifierTask(f, age)

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  protected def isCoveredFormula(f : Conjunction) : Boolean =
    ExQuantifierTask.isCoveredFormula(f)

}
