package ap.terfor.arithconj;

import scala.collection.mutable.ArrayBuffer

import ap.terfor.equations.{EquationConj, NegEquationConj, ReduceWithEqs}
import ap.util.{Debug, Logic}

object ArithConj {
  
  val AC = Debug.AC_PROP_CONNECTIVES
  
  def apply(positiveEqs : EquationConj, negativeEqs : NegEquationConj,
            order : TermOrder) : ArithConj = {
    if (positiveEqs.isFalse || negativeEqs.isFalse)
      FALSE
    else if (positiveEqs.isTrue && negativeEqs.isTrue)
      TRUE
    else
      new ArithConj (positiveEqs, negativeEqs, order)
  }
  
  def unapply(conj : ArithConj) : Option[(EquationConj, NegEquationConj)] =
    Some(conj.positiveEqs, conj.negativeEqs)
  
  val TRUE : ArithConj = new ArithConj(EquationConj.TRUE, NegEquationConj.TRUE,
                                       TermOrder.EMPTY)

  val FALSE : ArithConj = new ArithConj(EquationConj.FALSE, NegEquationConj.TRUE,
                                        TermOrder.EMPTY)
                                        
  /**
   * Compute the conjunction of equations, inequations and inequalities (later).
   */
  def conj(conjs : Iterable[Formula], order : TermOrder) : ArithConj = {
    val posEqs = new ArrayBuffer[EquationConj]
    val negEqs = new ArrayBuffer[NegEquationConj]
    
    for (conj <- conjs) {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPre(AC, conj.asInstanceOf[Sorted[Formula]] isSortedBy order)
      //////////////////////////////////////////////////////////////////////////
      
      conj match {
      case conj : EquationConj => posEqs += conj
      case conj : NegEquationConj => negEqs += conj
      case conj : ArithConj => {
        posEqs += conj.positiveEqs
        negEqs += conj.negativeEqs
      }
      }
    }
    
    apply(EquationConj.conj(posEqs, order),
          NegEquationConj.conj(negEqs, order),
          order)
  }

  def conj(f : Formula, order : TermOrder) : ArithConj = conj(Array(f), order)

}

/**
 * The class for a conjunction of equations, negated equations and inequalities
 */
class ArithConj private (val positiveEqs : EquationConj,
                         val negativeEqs : NegEquationConj,
                         val order : TermOrder)
                extends Formula with SortedWithOrder[ArithConj] {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(ArithConj.AC,
                   (positiveEqs isSortedBy order) &&
                   (negativeEqs isSortedBy order) &&
                   !negativeEqs.isFalse)
  //////////////////////////////////////////////////////////////////////////////

  def sortBy(newOrder : TermOrder) : ArithConj = {
    if (isSortedBy(newOrder)) {
      this
    } else {
      ArithConj (positiveEqs sortBy newOrder,
                 negativeEqs sortBy newOrder,
                 newOrder)
    }
  }
   
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Perform propagation between the different parts of the conjunction. Right
   * now, positive equations are propagated into negative equations
   */
  def internallyPropagate : ArithConj = 
    ArithConj (positiveEqs,
               ReduceWithEqs(positiveEqs, order)(negativeEqs),
               order)
    
  //////////////////////////////////////////////////////////////////////////////
    
  def isTrue : Boolean = (positiveEqs.isTrue && negativeEqs.isTrue)

  def isFalse : Boolean = (positiveEqs.isFalse || negativeEqs.isFalse)

  def size : Int = (positiveEqs.size + negativeEqs.size)
  
  /**
   * Return whether this conjunction actually only is a single literal
   */
  def isLiteral : Boolean = (this.size == 1)

  /**
   * Create the negation of at most one equation
   */
  def negate : ArithConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ArithConj.AC, this.size <= 1)
    ////////////////////////////////////////////////////////////////////////////
    if (positiveEqs.isEmpty) {
      ArithConj(negativeEqs.negate, NegEquationConj.TRUE, order)
    } else if (negativeEqs.isEmpty) {
      ArithConj(EquationConj.TRUE, positiveEqs.negate, order)
    } else {
      ArithConj.FALSE
    }
  }

  /**
   * Update the positive equations of this conjunction (without changing anything
   * else apart from the <code>TermOrder</code>) 
   */
  def updatePositiveEqs(newEqs : EquationConj, newOrder : TermOrder) : ArithConj =
    ArithConj(newEqs, this.negativeEqs, newOrder)
  
  /**
   * Update the Negative equations of this conjunction (without changing anything
   * else apart from the <code>TermOrder</code>) 
   */
  def updateNegativeEqs(newEqs : NegEquationConj, newOrder : TermOrder) : ArithConj =
    ArithConj(this.positiveEqs, newEqs, newOrder)
  
  //////////////////////////////////////////////////////////////////////////////

  lazy val variables : Set[VariableTerm] =
    positiveEqs.variables ++ negativeEqs.variables
    
  lazy val constants : Set[ConstantTerm] =
    positiveEqs.constants ++ negativeEqs.constants
      
  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = {
    if (isTrue) {
      "true"
    } else if (isFalse) {
      "false"
    } else {
      val strings = (for (lhs <- positiveEqs.elements) yield ("" + lhs + " = 0")) ++
                    (for (lhs <- negativeEqs.elements) yield ("" + lhs + " != 0"))
      if (strings.hasNext)
        strings.reduceLeft((s1 : String, s2 : String) =>
                           s1 + " & " + s2)
      else
        throw new Error // should never be reached
    }
  }

  override def equals(that : Any) : Boolean = that match {
    case that : ArithConj => this.positiveEqs == that.positiveEqs &&
                             this.negativeEqs == that.negativeEqs
    case _ => false
  }

  override def hashCode = (positiveEqs.hashCode + 726327 * negativeEqs.hashCode)
}
