package ap.terfor.arithconj;

import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj, ReduceWithEqs}
import ap.terfor.substitutions.VariableShiftSubst
import ap.util.{Debug, LazyMappedSet, UnionSet, Seqs, FilterIt}

object ReduceWithAC {
  
  private val AC = Debug.AC_PROPAGATION

  def apply(ac : ArithConj, order : TermOrder) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, ac isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    
    new ReduceWithAC(ReduceWithEqs(ac.positiveEqs, order),
                     ac.negativeEqs.toSet,
                     order)
  }
  
}

class ReduceWithAC private (positiveEqs : ReduceWithEqs,
                            negativeEqs : scala.collection.Set[LinearCombination],
                            order : TermOrder) {

  /**
   * Create a <code>ReduceWithEqs</code> that can be used underneath
   * <code>num</code> binders. The conversion of de Brujin-variables is done on
   * the fly, which should give a good performance when the resulting
   * <code>ReduceWithEqs</code> is not applied too often (TODO: caching)
   */
  private def passQuantifiers(num : Int) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, num >= 0)
    ////////////////////////////////////////////////////////////////////////////
    if (num == 0) {
      this
    } else {
      val shifters = VariableShiftSubst.upDownShifters(num, order)
      new ReduceWithAC(positiveEqs passQuantifiers num,
                       new LazyMappedSet(negativeEqs,
                                         shifters.upShifter[LinearCombination],
                                         shifters.downShifter[LinearCombination]),
                       order)
    }
  }

  private def addEquations(eqs : EquationConj) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    new ReduceWithAC(positiveEqs addEquations eqs.toMap,
                     negativeEqs,
                     order)
  }
  
  private def addEquations(eqs : NegEquationConj) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    new ReduceWithAC(positiveEqs,
                     new UnionSet (negativeEqs, eqs.toSet),
                     order)    
  }
  
  //////////////////////////////////////////////////////////////////////////////
  // Methods for reducing different kinds of formulas. If any of the
  // methods detects that a resulting formula is false,
  // <code>FALSE_EXCEPTION</code> is thrown (this is done to simplify the
  // handling of <code>Conjunction</code>s)
  // TODO: optimise for cases where no reduction is possible (detect this early,
  // create no new objects)
  
  private def reduce(eqs : EquationConj) : EquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redEqs = positiveEqs(eqs)
    if (redEqs.isFalse || !Seqs.disjoint(negativeEqs, redEqs))
      throw FALSE_EXCEPTION
    redEqs
  }

  private def reduce(eqs : NegEquationConj) : NegEquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redEqs = positiveEqs(eqs)
    if (redEqs.isFalse) throw FALSE_EXCEPTION
    NegEquationConj(FilterIt(redEqs.elements,
                             (lc:LinearCombination) => !(negativeEqs contains lc)),
                    order)
  }

  private def reduce(conj : Conjunction) : Conjunction = try {
    var reducer = this.passQuantifiers(conj.quans.size)
    
    val newAcPosEqs = reducer reduce conj.arithConj.positiveEqs
    reducer = reducer addEquations newAcPosEqs

    val newAcNegEqs = reducer reduce conj.arithConj.negativeEqs
    reducer = reducer addEquations newAcNegEqs

    val newArithConj = ArithConj(newAcPosEqs, newAcNegEqs, order)
    
    // TODO: later, also handle the other parts of a conjunction
    
    val newNegConjs = NegatedConjunctions(for (c <- conj.negatedConjs.elements)
                                          yield (reducer reduce c),
                                          order)
    
    val res = Conjunction(conj.quans, newArithConj, newNegConjs, order)
    
    if (newArithConj == res.arithConj) {
      res
    } else {
      // it might be necessary to repeat reduction, because new facts became
      // available
      this reduce res
    }
  } catch {
    case FALSE_EXCEPTION => Conjunction.FALSE
  }

  def apply(conj : Conjunction) : Conjunction = {
    val res = reduce(conj)
    ////////////////////////////////////////////////////////////////////////////
    // we demand that the reducer is a projection (repeated application does not
    // change the result anymore)
    Debug.assertPost(ReduceWithAC.AC, reduce(res) == res)
    ////////////////////////////////////////////////////////////////////////////
    res
  }
  
}

object FALSE_EXCEPTION extends Exception

