package ap.terfor.equations;

import ap.terfor.linearcombination.LinearCombination
import ap.basetypes.IdealInt
import ap.util.{Debug, Logic, FilterIt, Seqs}

object NegEquationConj {
  
  val AC = Debug.AC_EQUATIONS

  /**
   * Create an equation conjunction from an arbitrary set of equations
   * (left-hand sides).
   */
  def apply(lhss : Iterable[LinearCombination], order : TermOrder) : NegEquationConj =
    apply(lhss.elements, order)

  def apply(lhs : LinearCombination, order : TermOrder) : NegEquationConj =
    apply(Iterator.single(lhs), order)

  /**
   * Create an equation conjunction from an arbitrary set of equations
   * (left-hand sides).
   */
  def apply(lhss : Iterator[LinearCombination], order : TermOrder) : NegEquationConj =
    Seqs.filterAndSort[LinearCombination](lhss,
                                          l => l.isNonZero, l => l.isZero,
                                          l => l.makePrimitiveAndPositive,
                                          (a, b) => order.compare(a, b) > 0) match {
    case Seqs.FilteredSorted(sortedLhss) => {
      val contractedLhss = Seqs.removeDuplicates(sortedLhss).toArray
      new NegEquationConj (contractedLhss, order)
    }
    case Seqs.FoundBadElement => FALSE
    }

  val TRUE : NegEquationConj = new NegEquationConj (Array(), TermOrder.EMPTY)

  val FALSE : NegEquationConj = new NegEquationConj (Array(LinearCombination.ZERO),
                                                     TermOrder.EMPTY)


  /**
   * Compute the conjjunction of a number of conjunctions.
   * TODO: This could be optimised much more.
   */
  def conj(conjs : Iterable[NegEquationConj], order : TermOrder) : NegEquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, Logic.forall(for (c <- conjs.elements)
                                     yield (c isSortedBy order)))
    ////////////////////////////////////////////////////////////////////////////
      
    val lhss = for (d <- conjs.elements; lhs <- d.elements) yield lhs
      
    apply(lhss, order)
  }

}

/**
 * The class for representing disjunctions of equations (seen
 * as conjunctions of negated equations)
 */
class NegEquationConj private (_lhss : Array[LinearCombination],
                               _order : TermOrder)
      extends EquationSet(_lhss, _order) with SortedWithOrder[NegEquationConj] {

  def sortBy(newOrder : TermOrder) : NegEquationConj = {
    if (isSortedBy(newOrder))
      this
    else
      NegEquationConj(for (lc <- this.elements) yield lc.sortBy(newOrder),
                      newOrder)
  }

  //////////////////////////////////////////////////////////////////////////////

  def isTrue : Boolean = this.isEmpty

  /**
   * The only allowed case of obviously unsatisfiable conjunctions of negated
   * equations is the one of a single equation 0!=0
   */
  def isFalse : Boolean = (!isEmpty && this(0).isZero)

  /**
   * Create the negation of at most one negated equation
   */
  def negate : EquationConj = {
    if (this.isTrue) {
      EquationConj.FALSE
    } else {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPre(NegEquationConj.AC, this.size == 1)
      //////////////////////////////////////////////////////////////////////////
      EquationConj(this(0), order)
    }
  }

  protected val relationString : String = "!="

  //////////////////////////////////////////////////////////////////////////////
     
  override def equals(that : Any) : Boolean = that match {
    case that : NegEquationConj => super.equals(that)
    case _ => false
  }

  override def hashCode = (super.hashCode + 364783671)
}
